module cubic_mod

  implicit none

  contains

  subroutine quadratic(a,b,x) !x^2 + a*x + b = 0

    real,intent(in) ::  a,b
    complex,intent(inout) :: x(2)
    real :: discreminant

    discreminant = a*a - 4.*b
    if (discreminant >=0 ) then
      x(1) = cmplx((-a + sqrt(discreminant))*0.5,0.)
      x(2) = cmplx((-a - sqrt(discreminant))*0.5,0.)
    else
      x(1) = cmplx(-a*0.5,sqrt(-discreminant)*0.5)
      x(2) = cmplx(-a*0.5,-sqrt(-discreminant)*0.5)
    endif
    
  end subroutine   

  subroutine cubic_wroot(a,b,c,x0,x) !x^3 + a*x^2 + b*x + c = 0 knowing 1 real root x0
 
    real,intent(in) ::  a,b,c,x0
    complex,intent(inout) :: x(2)
    real :: p,q

    if (x0 == 0.) then
      call quadratic(a,b,x)
    else
      p = a + x0
      q = -c/x0
      call quadratic(p,q,x)
    endif

  end subroutine

  subroutine cubic01(a,b,c,d,x) ! a*x^3 + b*x^2 + c*x + d = 0 finds 1 real root between (0,1)

    real,intent(in) ::  a,b,c,d
    real,intent(inout) :: x
    complex :: xc(2)
    real :: a1,b1,c1

    a1 = 1.5*a + b
    b1 = c - 0.5625*a
    c1 = 0.03125*a + d

    if (a1 == 0.) then
      if (b1 /= 0) then 
        x = -c1/b1
      else
        call cubic_byNewton(-1.5,0.5625,d/a,0.,100,x,xc)
      endif
    else
        call quadratic(b1/a1,c1/a1,xc)
        if (real(xc(1)) >= 0 .and. real(xc(1)) <= 1) then
          x = real(xc(1))
        else
          x = real(xc(2))   
        endif
    endif

  end subroutine

  recursive subroutine cubic_byNewton(a,b,c,x0,n,x,xc) !x^3 + a*x^2 + b*x + c = 0 from initial guess x0

    real,intent(in) ::  a,b,c,x0
    integer,intent(in) :: n
    real,intent(inout) :: x
    complex,intent(inout):: xc(2)
    real :: f,fprime,x1
    integer :: i

    f = x0**3 +a*x0*x0 + b*x0 + c
    if (abs(f) <= 1e-8) then
      x = x0
      call cubic_wroot(a,b,c,x0,xc)
    else
      fprime = 3.*x0*x0 + 2.*a*x0 + b
      if (abs(fprime) <= 1e-8) then
        if (x0 == 0.) then
          x1 = 1.
        else
          x1 = 1.01*x0
        endif
        call cubic_byNewton(a,b,c,x1,n,x,xc)
      else
        x = x0
        do i=1,n
          f = x**3 +a*x*x + b*x + c
          if (abs(f) <= 1e-8) then
            exit
          else
            fprime = 3.*x*x + 2.*a*x + b
            x = x - f/fprime
          endif 
        enddo
        call cubic_wroot(a,b,c,x,xc)
      endif         
    endif

  end subroutine

  recursive subroutine cubic(a,b,c,x,xc)

    real,intent(in) ::  a,b,c
    real,intent(inout) :: x
    complex,intent(inout):: xc(2)
    real :: c1,alpha,b1,f1,a1,temp

    if (c == 0.) then
      x = 0.
      call quadratic(a,b,xc)
    elseif (c > 0) then
      call cubic(-a,b,-c,x,xc)
      x = -x
      xc = -xc
    else
      c1 = -c
      alpha = 1./(c1**(1./3.))
      a1 = alpha*a
      b1 = alpha*alpha*b
      f1 = a1 + b1
      if (f1 == 0.) then
        temp = 1.
      elseif (f1 > 0) then
        call cubic01(1.,a1,b1,-1.,temp)
      else
        call cubic01(-1.,b1,a1,1.,temp)
        temp = 1./temp
      endif
      temp = temp/alpha
      call cubic_byNewton(a,b,c,temp,100,x,xc) 
    endif

  end subroutine

end module
