module filter_mod

  use base_fftwv3_mod
  use ieee_arithmetic
  use ieee_exceptions
  use cubic_mod

  implicit none

  real, private :: dx,dy,dz,dt,ox,oy,oz,dx_shot,dy_shot,ox_shot,oy_shot,dx_rec,dy_rec,dz_rec,ox_rec,oy_rec,oz_rec,g(6),c10(0:5),c16(0:8),m,gama
  real, parameter :: pi = 4.*atan(1.)
  integer, private :: nx,ny,nz,nt,nx_pad,ny_pad,nz_pad_u,nz_pad_l,nx_shot,ny_shot,nx_rec,ny_rec,nz_rec,zshot,ratio,filter_length,half,halfx,halfy,halfz
  real, allocatable, private :: xtaper(:),ytaper(:),ztaper_u(:),ztaper_l(:),sou(:)
  real, allocatable :: ufilter(:,:,:),filter(:,:,:)
  integer, allocatable, private :: shot_loc(:,:),rec_loc(:,:)

  contains

  subroutine init_c16()
    ! Calculate coefficients of 16th order finite difference
    	  
    c16(1) = 1.
    c16(2) = -7./40.
    c16(3) = 7./165.
    c16(4) = -7./704.
    c16(5) = 7./3575.
    c16(6) = -1./3432.
    c16(7) = 1./35035.
    c16(8) = -1./732160.
    c16(0) = -2*sum(c16(1:8))
    c16 = 16./9.*c16
	  
  end subroutine

  subroutine init_c10()
    ! Calculate coefficients of 10th order finite difference

    c10(1) = 1.
    c10(2) = -1./7.
    c10(3) = 1./42.
    c10(4) = -1./336.
    c10(5) = 1./5250.
    c10(0) = -2*sum(c10(1:5))
    c10 = 5./3.*c10
	  
  end subroutine

  subroutine init_space(ox_in,oy_in,oz_in,dx_in,dy_in,dz_in,nx_in,ny_in,nz_in)

    real, intent(in) :: dx_in,dy_in,dz_in,ox_in,oy_in,oz_in
    integer, intent(in) :: nx_in,ny_in,nz_in
	
    dx = dx_in
    dy = dy_in
    dz = dz_in

    nx = nx_in
    ny = ny_in
    nz = nz_in

    ox = ox_in
    oy = oy_in
    oz = oz_in

  end subroutine  

  subroutine init_time(dt_in,nt_in,rate_in)
    real, intent(in) :: dt_in,rate_in
    integer, intent(in) :: nt_in
    dt = dt_in
    nt = nt_in
    ratio = nint(rate_in/dt)
  end subroutine

  subroutine init_shot(ox_shot_in,oy_shot_in,dx_shot_in,dy_shot_in,nx_shot_in,ny_shot_in,zshot_in)
  
    real, intent(in) :: dx_shot_in,dy_shot_in,ox_shot_in,oy_shot_in
    integer, intent(in) :: nx_shot_in,ny_shot_in,zshot_in
    integer :: ix_shot,iy_shot

    dx_shot = dx_shot_in
    dy_shot = dy_shot_in
    nx_shot = nx_shot_in
    ny_shot = ny_shot_in
    ox_shot = ox_shot_in
    oy_shot = oy_shot_in
    zshot = zshot_in

    allocate(shot_loc(nx_shot*ny_shot,2))  
    do iy_shot=1,ny_shot
      do ix_shot=1,nx_shot
        shot_loc((iy_shot-1)*nx_shot+ix_shot,1) = (ox_shot+(ix_shot-1)*dx_shot - ox)/dx + 1.5 
        shot_loc((iy_shot-1)*nx_shot+ix_shot,2) = (oy_shot+(iy_shot-1)*dy_shot - oy)/dy + 1.5 
      enddo
    enddo
  
  end subroutine 

  subroutine init_rec(ox_rec_in,oy_rec_in,oz_rec_in,dx_rec_in,dy_rec_in,dz_rec_in,nx_rec_in,ny_rec_in,nz_rec_in)
  
    real, intent(in) :: dx_rec_in,dy_rec_in,dz_rec_in,ox_rec_in,oy_rec_in,oz_rec_in
    integer, intent(in) :: nx_rec_in,ny_rec_in,nz_rec_in
    integer :: ix_rec,iy_rec,iz_rec
	
    dx_rec = dx_rec_in
    dy_rec = dy_rec_in
    dz_rec = dz_rec_in
    nx_rec = nx_rec_in
    ny_rec = ny_rec_in
    nz_rec = nz_rec_in
    ox_rec = ox_rec_in
    oy_rec = oy_rec_in
    oz_rec = oz_rec_in

    allocate(rec_loc(nx_rec*ny_rec*nz_rec,3))  
    do iz_rec=1,nz_rec
     do iy_rec=1,ny_rec
      do ix_rec=1,nx_rec
        rec_loc((iz_rec-1)*nx_rec*ny_rec+(iy_rec-1)*nx_rec+ix_rec,1) = (ox_rec+(ix_rec-1)*dx_rec - ox)/dx + 1.5 
        rec_loc((iz_rec-1)*nx_rec*ny_rec+(iy_rec-1)*nx_rec+ix_rec,2) = (oy_rec+(iy_rec-1)*dy_rec - oy)/dy + 1.5 
        rec_loc((iz_rec-1)*nx_rec*ny_rec+(iy_rec-1)*nx_rec+ix_rec,3) = (oz_rec+(iz_rec-1)*dz_rec - oz)/dz + 1.5 
      enddo
     enddo
    enddo
  
  end subroutine 

  subroutine init_source(tsou,ns)

    real, intent(in) :: tsou(:)
    integer, intent(in) :: ns

    allocate(sou(nt))
    sou = 0.
    sou(1:ns) = tsou

  end subroutine 

  subroutine init_taper(nx_pad_in,ny_pad_in,nz_pad_u_in,nz_pad_l_in,damper_in)
    real, intent(in) :: damper_in
    integer, intent(in) :: nx_pad_in,ny_pad_in,nz_pad_u_in,nz_pad_l_in
    real :: a0
    integer :: i

    nx_pad = nx_pad_in
    ny_pad = ny_pad_in
    nz_pad_u = nz_pad_u_in
    nz_pad_l = nz_pad_l_in
 
    allocate(ztaper_u(nz_pad_u),ztaper_l(nz_pad_l),xtaper(nx_pad),ytaper(ny_pad))

    a0 = damper_in/real(nz_pad_u)
    do i=1,nz_pad_u
      ztaper_u(i) = exp(-(a0*real(nz_pad_u+1-i))**2)
    enddo

    a0 = damper_in/real(nz_pad_l)
    do i=1,nz_pad_l
      ztaper_l(i) = exp(-(a0*real(nz_pad_l+1-i))**2)
    enddo

    a0 = damper_in/real(nx_pad)
    do i=1,nx_pad
      xtaper(i) = exp(-(a0*real(nx_pad+1-i))**2)
    enddo

    a0 = damper_in/real(ny_pad)
    do i=1,ny_pad
      ytaper(i) = exp(-(a0*real(ny_pad+1-i))**2)
    enddo

  end subroutine 

  subroutine taper(field)!      kz2 =
    real, intent(inout) :: field(nx,ny,nz)
    integer :: ix,iy,iz

    !$omp parallel do default(shared) private(ix,iy)
    do iy=1,ny
      do ix=1,nx 
        field(ix,iy,1:nz_pad_u) = field(ix,iy,1:nz_pad_u)*ztaper_u
        field(ix,iy,nz-nz_pad_l+1:nz) = field(ix,iy,nz-nz_pad_l+1:nz)*ztaper_l(nz_pad_l:1:-1)
      enddo
    enddo
    !$omp end parallel do

    !$omp parallel do default(shared) private(iz,iy)
    do iz=nz_pad_u+1,nz-nz_pad_l
      do iy=1,ny
        field(1:nx_pad,iy,iz) = field(1:nx_pad,iy,iz)*xtaper
        field(nx-nx_pad+1:nx,iy,iz) = field(nx-nx_pad+1:nx,iy,iz)*xtaper(nx_pad:1:-1)
      enddo
    enddo
    !$omp end parallel do

    !$omp parallel do default(shared) private(iz,ix)
    do iz=nz_pad_u+1,nz-nz_pad_l
      do ix=nx_pad+1,nx-nx_pad
        field(ix,1:ny_pad,iz) = field(ix,1:ny_pad,iz)*ytaper
        field(ix,ny-ny_pad+1:ny,iz) = field(ix,ny-ny_pad+1:ny,iz)*ytaper(ny_pad:1:-1)
      enddo
    enddo
    !$omp end parallel do

  end subroutine 

  subroutine clean_up()
    deallocate(shot_loc,rec_loc,sou,filter)
    if (allocated(xtaper)) deallocate(xtaper,ytaper,ztaper_u,ztaper_l)
    if (allocated(ufilter)) deallocate(ufilter)
  end subroutine 

  subroutine init_vel(eps1,eps2,del1,del2,del3,vpz)
  
    real, intent(in) :: eps1,eps2,del1,del2,del3,vpz

    g(6) = vpz*vpz
    g(1) = g(6)*(1.+2.*eps2)
    g(2) = g(1)*sqrt(1.+2.*del3)
    g(3) = g(6)*sqrt(1.+2.*del2)
    g(4) = g(6)*(1.+2.*eps1)
    g(5) = g(6)*sqrt(1.+2.*del1)

  end subroutine 

  subroutine init_filter(length,mode,exact)

    integer,intent(in) :: length,mode,exact
    real :: dkx,dky,dkz,kx2,ky2,kz2,t1,t2,p,s,detg,det1,det2,det3,a,b,c,x,err,kx,ky,kz,d,eig(3)
    complex :: xc(2)
    complex, allocatable :: cfilter_k(:,:,:)
    integer :: ix,iy,iz,temp(1)
    integer *8 :: plan

    call init_c10()
!    call init_c16()
    filter_length = length
    half = (filter_length-1)/2 

    halfx = (nx-1)/2 
    halfy = (ny-1)/2 
    halfz = (nz-1)/2 

    allocate(cfilter_k(nx-1,ny-1,nz-1),filter(nx-1,ny-1,nz-1))

    dkx = pi/dx/halfx
    dky = pi/dy/halfy
    dkz = pi/dz/halfz

    detg = g(1)*g(4)*g(6) + 2.*g(2)*g(3)*g(5) - g(1)*g(5)*g(5) - g(4)*g(3)*g(3) - g(6)*g(2)*g(2)
    det1 = g(4)*g(6) - g(5)*g(5)   
    det2 = g(1)*g(6) - g(3)*g(3)   
    det3 = g(4)*g(1) - g(2)*g(2)   

    filter = 0.
!    err = 0.
  if (exact == 1) then
    do iz=1,halfz+1 
      kz = (iz-1.)*dkz
!      kz2 = -2.*(c10(1)*(cos(kz*dz)-1.)+c10(2)*(cos(2.*kz*dz)-1.)+c10(3)*(cos(3.*kz*dz)-1.)+c10(4)*(cos(4.*kz*dz)-1.)+c10(5)*(cos(5.*kz*dz)-1.))/dz/dz
 !     kz2 = -2.*(c16(1)*(cos(kz*dz)-1.)+c16(2)*(cos(2.*kz*dz)-1.)+c16(3)*(cos(3.*kz*dz)-1.)+c16(4)*(cos(4.*kz*dz)-1.)+&
 !                      c16(5)*(cos(5.*kz*dz)-1.)+c16(6)*(cos(6.*kz*dz)-1.)+c16(7)*(cos(7.*kz*dz)-1.)+c16(8)*(cos(8.*kz*dz)-1.))/dz/dz
      kz2 = kz*kz
      do iy=1,halfy+1
        ky = (iy-1.)*dky
!        ky2 = -2.*(c10(1)*(cos(ky*dy)-1.)+c10(2)*(cos(2.*ky*dy)-1.)+c10(3)*(cos(3.*ky*dy)-1.)+c10(4)*(cos(4.*ky*dy)-1.)+c10(5)*(cos(5.*ky*dy)-1.))/dy/dy
 !       ky2 = -2.*(c16(1)*(cos(ky*dy)-1.)+c16(2)*(cos(2.*ky*dy)-1.)+c16(3)*(cos(3.*ky*dy)-1.)+c16(4)*(cos(4.*ky*dy)-1.)+&
 !                        c16(5)*(cos(5.*ky*dy)-1.)+c16(6)*(cos(6.*ky*dy)-1.)+c16(7)*(cos(7.*ky*dy)-1.)+c16(8)*(cos(8.*ky*dy)-1.))/dy/dy
        ky2 = ky*ky
        do ix=1,halfx+1  
          kx = (ix-1.)*dkx
!          kx2 = -2.*(c10(1)*(cos(kx*dx)-1.)+c10(2)*(cos(2.*kx*dx)-1.)+c10(3)*(cos(3.*kx*dx)-1.)+c10(4)*(cos(4.*kx*dx)-1.)+c10(5)*(cos(5.*kx*dx)-1.))/dx/dx
!          kx2 = -2.*(c16(1)*(cos(kx*dx)-1.)+c16(2)*(cos(2.*kx*dx)-1.)+c16(3)*(cos(3.*kx*dx)-1.)+c16(4)*(cos(4.*kx*dx)-1.)+&
!                           c16(5)*(cos(5.*kx*dx)-1.)+c16(6)*(cos(6.*kx*dx)-1.)+c16(7)*(cos(7.*kx*dx)-1.)+c16(8)*(cos(8.*kx*dx)-1.))/dx/dx
          kx2 = kx*kx

          a = g(1)*kx2 + g(4)*ky2 + g(6)*kz2
          b = kx2*ky2*det3 + ky2*kz2*det1 + kz2*kx2*det2
          c = detg*kx2*ky2*kz2
          call cubic(a,b,c,x,xc)
          eig = (/ x,real(xc(1)),real(xc(2)) /)
          call sort(eig)
!          write(0,*) x,real(xc(1)),real(xc(2))
          if (mode == 1) then
	    filter(ix,iy,iz) = eig(1)
!          write(0,*) filter(ix,iy,iz)
          elseif (mode == 2) then
	    filter(ix,iy,iz) = eig(2)
          else
	    filter(ix,iy,iz) = eig(3)
          endif

        enddo
      enddo
    enddo
  else
    do iz=1,halfz+1 
      kz = (iz-1.)*dkz
      kz2 = -2.*(c10(1)*(cos(kz*dz)-1.)+c10(2)*(cos(2.*kz*dz)-1.)+c10(3)*(cos(3.*kz*dz)-1.)+c10(4)*(cos(4.*kz*dz)-1.)+c10(5)*(cos(5.*kz*dz)-1.))/dz/dz
 !     kz2 = -2.*(c16(1)*(cos(kz*dz)-1.)+c16(2)*(cos(2.*kz*dz)-1.)+c16(3)*(cos(3.*kz*dz)-1.)+c16(4)*(cos(4.*kz*dz)-1.)+&
 !                      c16(5)*(cos(5.*kz*dz)-1.)+c16(6)*(cos(6.*kz*dz)-1.)+c16(7)*(cos(7.*kz*dz)-1.)+c16(8)*(cos(8.*kz*dz)-1.))/dz/dz
!      kz2 = kz*kz
      do iy=1,halfy+1
        ky = (iy-1.)*dky
        ky2 = -2.*(c10(1)*(cos(ky*dy)-1.)+c10(2)*(cos(2.*ky*dy)-1.)+c10(3)*(cos(3.*ky*dy)-1.)+c10(4)*(cos(4.*ky*dy)-1.)+c10(5)*(cos(5.*ky*dy)-1.))/dy/dy
 !       ky2 = -2.*(c16(1)*(cos(ky*dy)-1.)+c16(2)*(cos(2.*ky*dy)-1.)+c16(3)*(cos(3.*ky*dy)-1.)+c16(4)*(cos(4.*ky*dy)-1.)+&
 !                        c16(5)*(cos(5.*ky*dy)-1.)+c16(6)*(cos(6.*ky*dy)-1.)+c16(7)*(cos(7.*ky*dy)-1.)+c16(8)*(cos(8.*ky*dy)-1.))/dy/dy
 !       ky2 = ky*ky
        do ix=1,halfx+1  
          kx = (ix-1.)*dkx
          kx2 = -2.*(c10(1)*(cos(kx*dx)-1.)+c10(2)*(cos(2.*kx*dx)-1.)+c10(3)*(cos(3.*kx*dx)-1.)+c10(4)*(cos(4.*kx*dx)-1.)+c10(5)*(cos(5.*kx*dx)-1.))/dx/dx
!          kx2 = -2.*(c16(1)*(cos(kx*dx)-1.)+c16(2)*(cos(2.*kx*dx)-1.)+c16(3)*(cos(3.*kx*dx)-1.)+c16(4)*(cos(4.*kx*dx)-1.)+&
!                           c16(5)*(cos(5.*kx*dx)-1.)+c16(6)*(cos(6.*kx*dx)-1.)+c16(7)*(cos(7.*kx*dx)-1.)+c16(8)*(cos(8.*kx*dx)-1.))/dx/dx
!          kx2 = kx*kx

          a = g(1)*kx2 + g(4)*ky2 + g(6)*kz2
          b = kx2*ky2*det3 + ky2*kz2*det1 + kz2*kx2*det2
          c = detg*kx2*ky2*kz2
          call cubic(a,b,c,x,xc)
          eig = (/ x,real(xc(1)),real(xc(2)) /)
          call sort(eig)
!          write(0,*) x,real(xc(1)),real(xc(2))
          if (mode == 1) then
	    filter(ix,iy,iz) = eig(1)
!          write(0,*) filter(ix,iy,iz)
          elseif (mode == 2) then
	    filter(ix,iy,iz) = eig(2)
          else
	    filter(ix,iy,iz) = eig(3)
          endif

        enddo
      enddo
    enddo
  endif
 
   do iz=1,halfz+1
     do iy=1,halfy+1
      do ix=halfx+2,nx-1
        filter(ix,iy,iz) = filter(nx+1-ix,iy,iz)
      enddo
     enddo
     do iy=halfy+2,ny-1
      do ix=1,halfx+1
        filter(ix,iy,iz) = filter(ix,ny+1-iy,iz)
      enddo
     enddo
     do iy=halfy+2,ny-1
      do ix=halfx+2,nx-1
        filter(ix,iy,iz) = filter(nx+1-ix,ny+1-iy,iz)
      enddo
     enddo
    enddo

    do iz=halfz+2,nz-1
     do iy=1,ny-1
      do ix=1,nx-1
        filter(ix,iy,iz) = filter(ix,iy,nz+1-iz)
      enddo
     enddo
    enddo

    call sfftw_dft_3dx_init(plan, cfilter_k, nx-1, ny-1, nz-1, -1)  

    cfilter_k = cmplx(filter)
    call sfftw_dft_3dx_oper(plan, cfilter_k, nx-1, ny-1, nz-1)
    filter = real(cfilter_k)/sqrt((nx-1.)*(ny-1.)*(nz-1.))

    filter(1,1,1) = filter(1,1,1)*0.125
    filter(1,1,2:half+1) = filter(1,1,2:half+1)*0.25
    filter(1,2:half+1,1) = filter(1,2:half+1,1)*0.25
    filter(2:half+1,1,1) = filter(2:half+1,1,1)*0.25
    filter(1,2:half+1,2:half+1) = filter(1,2:half+1,2:half+1)*0.5
    filter(2:half+1,1,2:half+1) = filter(2:half+1,1,2:half+1)*0.5
    filter(2:half+1,2:half+1,1) = filter(2:half+1,2:half+1,1)*0.5

    deallocate(cfilter_k)

  end subroutine 

  subroutine write_filter()

    integer :: ix,iy,iz
    allocate(ufilter(nx,ny,nz))  

    ufilter(halfx+1:nx,halfy+1:ny,halfz+1:nz) = filter(1:halfx+1,1:halfy+1,1:halfz+1)

    do iz=halfz+1,nz
     do iy=halfy+1,ny
      do ix=1,halfx
        ufilter(ix,iy,iz) = ufilter(nx+1-ix,iy,iz)
      enddo
     enddo
     do iy=1,halfy
      do ix=halfx+1,nx
        ufilter(ix,iy,iz) = ufilter(ix,ny+1-iy,iz)
      enddo
     enddo
     do iy=1,halfy
      do ix=1,halfx
        ufilter(ix,iy,iz) = ufilter(nx+1-ix,ny+1-iy,iz)
      enddo
     enddo
    enddo

    do iz=1,halfz
     do iy=1,ny
      do ix=1,nx
        ufilter(ix,iy,iz) = ufilter(ix,iy,nz+1-iz)
      enddo
     enddo
    enddo  

  end subroutine 

  subroutine synthesize(dat)

    real, intent(inout) :: dat(:,:,:,:,:,:)
    integer:: ix_shot,iy_shot,it,ishot,ix_rec,iy_rec,iz_rec,irec
    real, allocatable :: cur(:,:,:),prev(:,:,:),next(:,:,:),temp(:,:,:)

    allocate(cur(nx,ny,nz), prev(nx,ny,nz), next(nx,ny,nz), temp(nx,ny,nz))

    cur = 0.
    prev = 0.
    next = 0.
    temp = 0.

    do iy_shot=1,ny_shot
     do ix_shot=1,nx_shot
      ishot=(iy_shot-1)*nx_shot+ix_shot
      do it=1,nt
        write(0,*) it

        cur(shot_loc(ishot,1),shot_loc(ishot,2),zshot) = cur(shot_loc(ishot,1),shot_loc(ishot,2),zshot) + sou(it) 

        call partial_space2(cur,temp)
        next = temp*dt*dt + 2.*cur - prev

        if (maxval(abs(next)) > 1.e17) then
          write(0,*) "blow up"
          write(0,*) maxval(abs(next))
          exit  ! Exit the loop when the amplitude blows up
        endif

        if (.not. ieee_is_finite(sum(next))) then
          write(0,*) "nan"
          exit ! Exit the loop when NaN is found
        endif

!        call taper(cur)
!        call taper(next)

        if (mod(it-1,ratio) == 0) then
          do iz_rec=1,nz_rec
           do iy_rec=1,ny_rec
            do ix_rec=1,nx_rec
              irec=(iz_rec-1)*nx_rec*ny_rec+(iy_rec-1)*nx_rec+ix_rec
              dat((it-1)/ratio+1,ix_rec,iy_rec,iz_rec,ix_shot,iy_shot)  = cur(rec_loc(irec,1),rec_loc(irec,2),rec_loc(irec,3))
            enddo
           enddo
          enddo
        endif

        prev = cur ! Update wavefield
        cur = next

      enddo ! end time loop

     enddo ! end shot loop
    enddo
   
    deallocate(prev,cur,next,temp)

  end subroutine 

  subroutine partial_space2(p,dp)

    real, intent(in) :: p(nx,ny,nz)
    real, intent(inout) :: dp(nx,ny,nz)
    integer :: ix,iy,iz,ifx,ify,ifz

    !$omp parallel do default(shared) private(ix,iy,iz,ifx,ify,ifz)
    do iz=half+1,nz-half 
      do iy=half+1,ny-half
        do ix=half+1,nx-half

          dp(ix,iy,iz) = 0.
          do ifz=1,half+1
            do ify=1,half+1
             do ifx=1,half+1
                  dp(ix,iy,iz) = dp(ix,iy,iz) + filter(ifx,ify,ifz)*(p(ix+ifx-1,iy+ify-1,iz+ifz-1)+&
                                                                                       p(ix+ifx-1,iy+ify-1,iz-ifz+1)+&
                                                                                       p(ix+ifx-1,iy-ify+1,iz+ifz-1)+&
                                                                                       p(ix-ifx+1,iy+ify-1,iz+ifz-1)+&
                                                                                       p(ix+ifx-1,iy-ify+1,iz-ifz+1)+&
                                                                                       p(ix-ifx+1,iy+ify-1,iz-ifz+1)+&
                                                                                       p(ix-ifx+1,iy-ify+1,iz+ifz-1)+&
                                                                                       p(ix-ifx+1,iy-ify+1,iz-ifz+1))
              enddo
            enddo            
          enddo

        enddo
      enddo
    enddo
    !$omp end parallel do

  end subroutine 

  subroutine sort(ARR)
    real,intent(inout) :: ARR(3)
    real :: a  
    integer :: i,j
    do j=2, 3
      a=ARR(j)
      do i=j-1,1,-1
      if (ARR(i)<=a) goto 10
        ARR(i+1)=ARR(i)
      end do
	i=0
10  ARR(i+1)=a
    end do
 
  end subroutine

end module
