module my_mod

  use ieee_arithmetic
  use ieee_exceptions

  implicit none
  
  real, private :: dx,dy,dz,dt,ox,oy,oz,dx_shot,dy_shot,ox_shot,oy_shot,dx_rec,dy_rec,dz_rec,ox_rec,oy_rec,oz_rec
  real, parameter :: pi = 4.*atan(1.)
  integer, private :: nx,ny,nz,nt,nx_pad,ny_pad,nz_pad_u,nz_pad_l,nx_shot,ny_shot,nx_rec,ny_rec,nz_rec,zshot,ratio
  real, allocatable, private :: xtaper(:),ytaper(:),ztaper_u(:),ztaper_l(:),sou(:)
  real, allocatable :: m(:,:,:,:),dm(:,:,:,:)
  real, private :: c8(0:4),dx_2,dy_2,dz_2,b8(4)
  integer, allocatable, private :: shot_loc(:,:),rec_loc(:,:)

  contains 

  subroutine init_b8()
    ! Calculate coefficients of 8th order finite difference

    b8(1) = 1.
    b8(2) = -1./4.
    b8(3) = 1./21.
    b8(4) = -1./224.
    b8 = 4./5.*b8
	  
  end subroutine

  subroutine init_c8()

    c8(1) = 1.
    c8(2) = -1./8.
    c8(3) = 1./63.
    c8(4) = -1./896.
    c8(0) = -2*sum(c8(1:4))
    c8 = 8./5.*c8
	  
  end subroutine
  
  subroutine init_space(ox_in,oy_in,oz_in,dx_in,dy_in,dz_in,nx_in,ny_in,nz_in)

    real, intent(in) :: dx_in,dy_in,dz_in,ox_in,oy_in,oz_in
    integer, intent(in) :: nx_in,ny_in,nz_in
	
    dx = dx_in
    dy = dy_in
    dz = dz_in

    nx = nx_in
    ny = ny_in
    nz = nz_in

    ox = ox_in
    oy = oy_in
    oz = oz_in

    dx_2 = 1./dx/dx
    dy_2 = 1./dy/dy
    dz_2 = 1./dz/dz

  end subroutine  

  subroutine init_time(dt_in,nt_in,rate_in)
    real, intent(in) :: dt_in,rate_in
    integer, intent(in) :: nt_in
    dt = dt_in
    nt = nt_in
    ratio = nint(rate_in/dt)
  end subroutine

  subroutine init_shot(ox_shot_in,oy_shot_in,dx_shot_in,dy_shot_in,nx_shot_in,ny_shot_in,zshot_in)
  
    real, intent(in) :: dx_shot_in,dy_shot_in,ox_shot_in,oy_shot_in
    integer, intent(in) :: nx_shot_in,ny_shot_in,zshot_in
    integer :: ix_shot,iy_shot

    dx_shot = dx_shot_in
    dy_shot = dy_shot_in
    nx_shot = nx_shot_in
    ny_shot = ny_shot_in
    ox_shot = ox_shot_in
    oy_shot = oy_shot_in
    zshot = zshot_in

    allocate(shot_loc(nx_shot*ny_shot,2))  
    do iy_shot=1,ny_shot
      do ix_shot=1,nx_shot
        shot_loc((iy_shot-1)*nx_shot+ix_shot,1) = (ox_shot+(ix_shot-1)*dx_shot - ox)/dx + 1.5 
        shot_loc((iy_shot-1)*nx_shot+ix_shot,2) = (oy_shot+(iy_shot-1)*dy_shot - oy)/dy + 1.5 
      enddo
    enddo
  
  end subroutine 

  subroutine init_rec(ox_rec_in,oy_rec_in,oz_rec_in,dx_rec_in,dy_rec_in,dz_rec_in,nx_rec_in,ny_rec_in,nz_rec_in)
  
    real, intent(in) :: dx_rec_in,dy_rec_in,dz_rec_in,ox_rec_in,oy_rec_in,oz_rec_in
    integer, intent(in) :: nx_rec_in,ny_rec_in,nz_rec_in
    integer :: ix_rec,iy_rec,iz_rec
	
    dx_rec = dx_rec_in
    dy_rec = dy_rec_in
    dz_rec = dz_rec_in
    nx_rec = nx_rec_in
    ny_rec = ny_rec_in
    nz_rec = nz_rec_in
    ox_rec = ox_rec_in
    oy_rec = oy_rec_in
    oz_rec = oz_rec_in

    allocate(rec_loc(nx_rec*ny_rec*nz_rec,3))  
    do iz_rec=1,nz_rec
     do iy_rec=1,ny_rec
      do ix_rec=1,nx_rec
        rec_loc((iz_rec-1)*nx_rec*ny_rec+(iy_rec-1)*nx_rec+ix_rec,1) = (ox_rec+(ix_rec-1)*dx_rec - ox)/dx + 1.5 
        rec_loc((iz_rec-1)*nx_rec*ny_rec+(iy_rec-1)*nx_rec+ix_rec,2) = (oy_rec+(iy_rec-1)*dy_rec - oy)/dy + 1.5 
        rec_loc((iz_rec-1)*nx_rec*ny_rec+(iy_rec-1)*nx_rec+ix_rec,3) = (oz_rec+(iz_rec-1)*dz_rec - oz)/dz + 1.5 
      enddo
     enddo
    enddo
  
  end subroutine 

  subroutine init_source(tsou,ns)

    real, intent(in) :: tsou(:)
    integer, intent(in) :: ns

    allocate(sou(nt))
    sou = 0.
    sou(1:ns) = tsou

  end subroutine 
  
  subroutine synthesize2(dat)

    real, intent(inout) :: dat(:,:,:,:,:,:)
    integer:: ix_shot,iy_shot,it,ishot,ix_rec,iy_rec,iz_rec,irec
    real, allocatable :: cur(:,:,:,:),prev(:,:,:,:),next(:,:,:,:),temp(:,:,:,:)

    allocate(cur(nx,ny,nz,3), prev(nx,ny,nz,3), next(nx,ny,nz,3), temp(nx,ny,nz,3))
    call init_c8()

    cur = 0.
    prev = 0.
    next = 0.
    temp = 0.

    do iy_shot=1,ny_shot
     do ix_shot=1,nx_shot
      ishot=(iy_shot-1)*nx_shot+ix_shot
      do it=1,nt
        write(0,*) it

        cur(shot_loc(ishot,1),shot_loc(ishot,2),zshot,:) = cur(shot_loc(ishot,1),shot_loc(ishot,2),zshot,:) + sou(it) 

        call partial_space2(cur,temp)
        call sym_matmul(m,temp,next,.false.)
        next = next*dt*dt + 2.*cur - prev

        if (maxval(abs(next)) > 1.e17) then
          write(0,*) "blow up"
          write(0,*) maxval(abs(next))
          exit  ! Exit the loop when the amplitude blows up
        endif

        if (.not. ieee_is_finite(sum(next))) then
          write(0,*) "nan"
          exit ! Exit the loop when NaN is found
        endif

!        call taper(cur)
!        call taper(next)

        temp(:,:,:,1) = - 1./3.*sum(cur,4)
        if (mod(it-1,ratio) == 0) then
          do iz_rec=1,nz_rec
           do iy_rec=1,ny_rec
            do ix_rec=1,nx_rec
              irec=(iz_rec-1)*nx_rec*ny_rec+(iy_rec-1)*nx_rec+ix_rec
              dat((it-1)/ratio+1,ix_rec,iy_rec,iz_rec,ix_shot,iy_shot)  = temp(rec_loc(irec,1),rec_loc(irec,2),rec_loc(irec,3),1)
            enddo
           enddo
          enddo
        endif

        prev = cur ! Update wavefield
        cur = next

      enddo ! end time loop

     enddo ! end shot loop
    enddo
   
    deallocate(prev,cur,next,temp)

  end subroutine 

  subroutine synthesize1(dat)

    real, intent(inout) :: dat(:,:,:,:,:,:)
    integer:: ix_shot,iy_shot,it,ishot,ix_rec,iy_rec,iz_rec,irec
    real, allocatable :: cur(:,:,:,:),prev(:,:,:,:),next(:,:,:,:),temp(:,:,:,:)

    allocate(cur(nx,ny,nz,3), prev(nx,ny,nz,3), next(nx,ny,nz,3), temp(nx,ny,nz,3))
    call init_b8()

    cur = 0.
    prev = 0.
    next = 0.
    temp = 0.

    do iy_shot=1,ny_shot
     do ix_shot=1,nx_shot
      ishot=(iy_shot-1)*nx_shot+ix_shot
      do it=1,nt
        write(0,*) it

        cur(shot_loc(ishot,1),shot_loc(ishot,2),zshot,:) = cur(shot_loc(ishot,1),shot_loc(ishot,2),zshot,:) + sou(it) 

        call partial_space1(cur,next)
        call sym_matmul(m,next,temp,.false.)
        call partial_space1(temp,next)
        next = next*dt*dt + 2.*cur - prev

        if (maxval(abs(next)) > 1.e17) then
          write(0,*) "blow up"
          write(0,*) maxval(abs(next))
          exit  ! Exit the loop when the amplitude blows up
        endif

        if (.not. ieee_is_finite(sum(next))) then
          write(0,*) "nan"
          exit ! Exit the loop when NaN is found
        endif

!        call taper(cur)
!        call taper(next)

        temp(:,:,:,1) = - 1./3.*sum(cur,4)
        if (mod(it-1,ratio) == 0) then
          do iz_rec=1,nz_rec
           do iy_rec=1,ny_rec
            do ix_rec=1,nx_rec
              irec=(iz_rec-1)*nx_rec*ny_rec+(iy_rec-1)*nx_rec+ix_rec
              dat((it-1)/ratio+1,ix_rec,iy_rec,iz_rec,ix_shot,iy_shot)  = temp(rec_loc(irec,1),rec_loc(irec,2),rec_loc(irec,3),1)
            enddo
           enddo
          enddo
        endif

        prev = cur ! Update wavefield
        cur = next

      enddo ! end time loop

     enddo ! end shot loop
    enddo
   
    deallocate(prev,cur,next,temp)

  end subroutine 

  subroutine perturb_data(dat)

    real, intent(inout) :: dat(:,:,:,:,:,:)
    integer:: ix_shot,iy_shot,it,ishot,ix_rec,iy_rec,iz_rec,irec
    real, allocatable :: cur(:,:,:,:),prev(:,:,:,:),next(:,:,:,:),temp(:,:,:,:),dcur(:,:,:,:),dprev(:,:,:,:),dnext(:,:,:,:)

    allocate(cur(nx,ny,nz,3), prev(nx,ny,nz,3), next(nx,ny,nz,3), temp(nx,ny,nz,3))
    allocate(dcur(nx,ny,nz,3), dprev(nx,ny,nz,3), dnext(nx,ny,nz,3))
    cur = 0.
    prev = 0.
    next = 0.
    dcur = 0.
    dprev = 0.
    dnext = 0.
    temp = 0.

    do iy_shot=1,ny_shot
     do ix_shot=1,nx_shot
      ishot=(iy_shot-1)*nx_shot+ix_shot
      do it=1,nt
        write(0,*) it

        cur(shot_loc(ishot,1),shot_loc(ishot,2),zshot,:) = cur(shot_loc(ishot,1),shot_loc(ishot,2),zshot,:) + sou(it) 
        call partial_space2(cur,temp)
        call sym_matmul(m,temp,next,.false.)
        next = next*dt*dt + 2.*cur - prev

        call sym_matmul(dm,temp,dnext,.false.)
        call partial_space2(dcur,temp)
        call sym_matmul(m,temp,dnext,.true.)
        dnext = dnext*dt*dt + 2.*dcur - dprev

        if (maxval(abs(next)) > 1.e17) then
          write(0,*) "blow up in background"
          write(0,*) maxval(abs(next))
          exit  ! Exit the loop when the amplitude blows up
        endif

        if (.not. ieee_is_finite(sum(next))) then
          write(0,*) "nan in background"
          exit ! Exit the loop when NaN is found
        endif

        if (maxval(abs(dnext)) > 1.e17) then
          write(0,*) "blow up in perturbation"
          write(0,*) maxval(abs(dnext))
          exit  ! Exit the loop when the amplitude blows up
        endif

        if (.not. ieee_is_finite(sum(dnext))) then
          write(0,*) "nan in perturbation"
          exit ! Exit the loop when NaN is found
        endif
        
        call taper(cur)
        call taper(next)
        call taper(dcur)
        call taper(dnext)

        temp(:,:,:,1) = - 1./3.*sum(dcur,4)
        if (mod(it-1,ratio) == 0) then
          do iz_rec=1,nz_rec
           do iy_rec=1,ny_rec
            do ix_rec=1,nx_rec
              irec=(iz_rec-1)*nx_rec*ny_rec+(iy_rec-1)*nx_rec+ix_rec
              dat((it-1)/ratio+1,ix_rec,iy_rec,iz_rec,ix_shot,iy_shot)  = temp(rec_loc(irec,1),rec_loc(irec,2),rec_loc(irec,3),1)
            enddo
           enddo
          enddo
        endif

        prev = cur ! Update wavefield
        cur = next
        dprev = dcur ! Update wavefield
        dcur = dnext

      enddo ! end time loop

     enddo ! end shot loop
    enddo
   
    deallocate(prev,cur,next,temp,dprev,dcur,dnext)

  end subroutine 

  subroutine init_model(eps1,eps2,del1,del2,del3,vpz)
  
    real, intent(in) :: eps1(nx,ny,nz),eps2(nx,ny,nz),del1(nx,ny,nz),del2(nx,ny,nz),del3(nx,ny,nz),vpz(nx,ny,nz)

    allocate(m(nx,ny,nz,6))

    m(:,:,:,6) = vpz*vpz
    m(:,:,:,1) = m(:,:,:,6)*(1.+2.*eps2)
    m(:,:,:,2) = m(:,:,:,1)*sqrt(1.+2.*del3)
    m(:,:,:,3) = m(:,:,:,6)*sqrt(1.+2.*del2)
    m(:,:,:,4) = m(:,:,:,6)*(1.+2.*eps1)
    m(:,:,:,5) = m(:,:,:,6)*sqrt(1.+2.*del1)
    write(0,*) maxval(m), minval(m)
  end subroutine 

  subroutine perturb_model(a,perc)
  
    real, intent(in) :: perc(6)
    integer, intent(in) :: a(6)
    integer :: i

    allocate(dm(nx,ny,nz,6))
    dm = 0.
    do i=1,6
      if (a(i) .eq. 1) dm((nx+1)/2-4:(nx+1)/2+4,(ny+1)/2-4:(ny+1)/2+4,(nz+1)/2-4:(nz+1)/2+4,i) = perc(i)*m((nx+1)/2-4:(nx+1)/2+4,(ny+1)/2-4:(ny+1)/2+4,(nz+1)/2-4:(nz+1)/2+4,i)*0.01
    enddo
    write(0,*) maxval(dm), minval(dm)
  end subroutine 

  subroutine update_model()
    m = m + dm
    write(0,*) maxval(m), minval(m)
  end subroutine

  subroutine vel2thomsen(eps1,eps2,del1,del2,del3,vpz)
  
    real, intent(inout) :: eps1(nx,ny,nz),eps2(nx,ny,nz),del1(nx,ny,nz),del2(nx,ny,nz),del3(nx,ny,nz),vpz(nx,ny,nz)
  
    vpz = sqrt(m(:,:,:,6))
    eps2 = (m(:,:,:,1)/m(:,:,:,6) - 1.)/2.
    del3 = ((m(:,:,:,2)/m(:,:,:,1))**2 - 1.)/2
    del2 = ((m(:,:,:,3)/m(:,:,:,6))**2 -1.)/2.
    eps1 = (m(:,:,:,4)/m(:,:,:,6) - 1.)/2.
    del1 = ((m(:,:,:,5)/m(:,:,:,6))**2 -1.)/2.

  end subroutine

  subroutine partial_space2(p,dp)

    real, intent(in) :: p(nx,ny,nz,3)
    real, intent(inout) :: dp(nx,ny,nz,3)
    integer :: ix,iy,iz,i,j,k

    !$omp parallel do default(shared) private(ix,iy,iz)
    do iz=5,nz-4 
      do iy=5,ny-4
        do ix=5,nx-4

          dp(ix,iy,iz,1) = dx_2*(c8(0)*p(ix,iy,iz,1) &
                                           + c8(1)*(p(ix+1,iy,iz,1) + p(ix-1,iy,iz,1)) &
                                           + c8(2)*(p(ix+2,iy,iz,1) + p(ix-2,iy,iz,1)) &
                                           + c8(3)*(p(ix+3,iy,iz,1) + p(ix-3,iy,iz,1)) &
                                           + c8(4)*(p(ix+4,iy,iz,1) + p(ix-4,iy,iz,1)))

          dp(ix,iy,iz,2) = dy_2*(c8(0)*p(ix,iy,iz,2) &
                                           + c8(1)*(p(ix,iy+1,iz,2) + p(ix,iy-1,iz,2)) &
                                           + c8(2)*(p(ix,iy+2,iz,2) + p(ix,iy-2,iz,2)) &
                                           + c8(3)*(p(ix,iy+3,iz,2) + p(ix,iy-3,iz,2)) &
                                           + c8(4)*(p(ix,iy+4,iz,2) + p(ix,iy-4,iz,2)))
 
         dp(ix,iy,iz,3) = dz_2*(c8(0)*p(ix,iy,iz,3) &
                                           + c8(1)*(p(ix,iy,iz+1,3) + p(ix,iy,iz-1,3)) &
                                           + c8(2)*(p(ix,iy,iz+2,3) + p(ix,iy,iz-2,3)) &
                                           + c8(3)*(p(ix,iy,iz+3,3) + p(ix,iy,iz-3,3)) &
                                           + c8(4)*(p(ix,iy,iz+4,3) + p(ix,iy,iz-4,3)))
        enddo
      enddo
    enddo
    !$omp end parallel do

  end subroutine    

  subroutine partial_space1(p,dp)

    real, intent(in) :: p(nx,ny,nz,3)
    real, intent(inout) :: dp(nx,ny,nz,3)
    integer :: ix,iy,iz,i,j,k

    !$omp parallel do default(shared) private(ix,iy,iz)
    do iz=5,nz-4 
      do iy=5,ny-4
        do ix=5,nx-4

          dp(ix,iy,iz,1) = (b8(1)*(p(ix+1,iy,iz,1) - p(ix-1,iy,iz,1)) &
                                + b8(2)*(p(ix+2,iy,iz,1) - p(ix-2,iy,iz,1)) &
                                + b8(3)*(p(ix+3,iy,iz,1) - p(ix-3,iy,iz,1)) &
                                + b8(4)*(p(ix+4,iy,iz,1) - p(ix-4,iy,iz,1)))/dx

          dp(ix,iy,iz,2) = (b8(1)*(p(ix,iy+1,iz,2) - p(ix,iy-1,iz,2)) &
                                + b8(2)*(p(ix,iy+2,iz,2) - p(ix,iy-2,iz,2)) &
                                + b8(3)*(p(ix,iy+3,iz,2) - p(ix,iy-3,iz,2)) &
                                + b8(4)*(p(ix,iy+4,iz,2) - p(ix,iy-4,iz,2)))/dy
 
         dp(ix,iy,iz,3) = (b8(1)*(p(ix,iy,iz+1,3) - p(ix,iy,iz-1,3)) &
                               + b8(2)*(p(ix,iy,iz+2,3) - p(ix,iy,iz-2,3)) &
                               + b8(3)*(p(ix,iy,iz+3,3) - p(ix,iy,iz-3,3)) &
                               + b8(4)*(p(ix,iy,iz+4,3) - p(ix,iy,iz-4,3)))/dz
        enddo
      enddo
    enddo
    !$omp end parallel do

  end subroutine    

  subroutine sym_matmul(a,x,y,add)

    real, intent(in) :: a(nx,ny,nz,6)
    real, intent(in) :: x(nx,ny,nz,3)
    logical, intent(in) :: add
    real, intent(inout) :: y(nx,ny,nz,3)

    if (.not. add) y = 0.
    y(:,:,:,1) = y(:,:,:,1) + a(:,:,:,1)*x(:,:,:,1) + a(:,:,:,2)*x(:,:,:,2) + a(:,:,:,3)*x(:,:,:,3)
    y(:,:,:,2) = y(:,:,:,2) + a(:,:,:,2)*x(:,:,:,1) + a(:,:,:,4)*x(:,:,:,2) + a(:,:,:,5)*x(:,:,:,3)
    y(:,:,:,3) = y(:,:,:,3) + a(:,:,:,3)*x(:,:,:,1) + a(:,:,:,5)*x(:,:,:,2) + a(:,:,:,6)*x(:,:,:,3)
    
  end subroutine

  subroutine init_taper(nx_pad_in,ny_pad_in,nz_pad_u_in,nz_pad_l_in,damper_in)
    real, intent(in) :: damper_in
    integer, intent(in) :: nx_pad_in,ny_pad_in,nz_pad_u_in,nz_pad_l_in
    real :: a0
    integer :: i

    nx_pad = nx_pad_in
    ny_pad = ny_pad_in
    nz_pad_u = nz_pad_u_in
    nz_pad_l = nz_pad_l_in
 
    allocate(ztaper_u(nz_pad_u),ztaper_l(nz_pad_l),xtaper(nx_pad),ytaper(ny_pad))

    a0 = damper_in/real(nz_pad_u)
    do i=1,nz_pad_u
      ztaper_u(i) = exp(-(a0*real(nz_pad_u+1-i))**2)
    enddo

    a0 = damper_in/real(nz_pad_l)
    do i=1,nz_pad_l
      ztaper_l(i) = exp(-(a0*real(nz_pad_l+1-i))**2)
    enddo

    a0 = damper_in/real(nx_pad)
    do i=1,nx_pad
      xtaper(i) = exp(-(a0*real(nx_pad+1-i))**2)
    enddo

    a0 = damper_in/real(ny_pad)
    do i=1,ny_pad
      ytaper(i) = exp(-(a0*real(ny_pad+1-i))**2)
    enddo

  end subroutine 

  subroutine taper(field)
    real, intent(inout) :: field(nx,ny,nz,3)
    integer :: ix,iy,iz,i

    do i=1,3
    !$omp parallel do default(shared) private(ix,iy)
    do iy=1,ny
      do ix=1,nx 
        field(ix,iy,1:nz_pad_u,i) = field(ix,iy,1:nz_pad_u,i)*ztaper_u
        field(ix,iy,nz-nz_pad_l+1:nz,i) = field(ix,iy,nz-nz_pad_l+1:nz,i)*ztaper_l(nz_pad_l:1:-1)
      enddo
    enddo
    !$omp end parallel do

    !$omp parallel do default(shared) private(iz,iy)
    do iz=nz_pad_u+1,nz-nz_pad_l
      do iy=1,ny
        field(1:nx_pad,iy,iz,i) = field(1:nx_pad,iy,iz,i)*xtaper
        field(nx-nx_pad+1:nx,iy,iz,i) = field(nx-nx_pad+1:nx,iy,iz,i)*xtaper(nx_pad:1:-1)
      enddo
    enddo
    !$omp end parallel do

    !$omp parallel do default(shared) private(iz,ix)
    do iz=nz_pad_u+1,nz-nz_pad_l
      do ix=nx_pad+1,nx-nx_pad
        field(ix,1:ny_pad,iz,i) = field(ix,1:ny_pad,iz,i)*ytaper
        field(ix,ny-ny_pad+1:ny,iz,i) = field(ix,ny-ny_pad+1:ny,iz,i)*ytaper(ny_pad:1:-1)
      enddo
    enddo
    !$omp end parallel do

    enddo

  end subroutine 

  subroutine clean_up()
    deallocate(m,shot_loc,rec_loc,sou)
    if (allocated(dm)) deallocate(dm)
    if (allocated(xtaper)) deallocate(xtaper,ytaper,ztaper_u,ztaper_l)
  end subroutine 
    
end module

